
/** \file stl_accutimer.h
 *		This file contains a timer class that runs at a
 *		resolution of 4 microseconds. It is intended to
 *		keep time with a higher degree of accuracy than
 *		the original stl_timer class by changing the
 *		underlying mechanism by which interrupts are
 *		triggered.
 *
 *		This class shall serve as a drop-in replacement
 *		for the stl_timer after it goes through enough
 *		iterations to be considered bug-free and stable.
 */

#ifndef _STL_ACCUTIMER_H
#define _STL_ACCUTIMER_H

#ifndef F_CPU
  #error Must have the macro F_CPU defined in make!
#endif

#ifdef TCNT3
  #define TMR_TCNT       TCNT3
  #define TMR_OVF_vect   TIMER3_OVF_vect
  #define TMR_CAPT_vect  TIMER3_CAPT_vect
  #define TMR_COMPA_vect TIMER3_COMPA_vect
  #define TMR_COMPB_vect TIMER3_COMPB_vect
  #define TMR_COMPC_vect TIMER3_COMPC_vect
#else
  #define TMR_TCNT       TCNT1
  #define TMR_OVF_vect   TIMER1_OVF_vect
  #define TMR_CAPT_vect  TIMER1_CAPT_vect
  #define TMR_COMPA_vect TIMER1_COMPA_vect
  #define TMR_COMPB_vect TIMER1_COMPB_vect
  #define TMR_COMPC_vect TIMER1_COMPC_vect
#endif

#define QSECS_PER_MIN         (240)
#define QSECS_PER_MIN_64      (240ULL)
#define MIN_PER_DAY           (1440)
#define MIN_PER_DAY_64        (1440ULL)
#define TICKS_PER_QSEC        (62500)
#define TICKS_PER_QSEC_64     (62500ULL)
#define TICKS_PER_MIN_64      (15000000ULL) //62.5K * 4 * 60 = 15M
#define TICKS_PER_DAY_64      (21600000000ULL) //15M * 60 * 24 = 21.6G

typedef struct
{
  uint16_t ticks;
  uint8_t qSecs;
  uint16_t minutes;
} ClockData;

/**
 * Converts a ClockData struct into its representative total tick count.
 */
uint64_t convertClockData(ClockData clk_data)
{
  uint64_t totalTicks = clk_data.minutes * TICKS_PER_MIN_64;
  totalTicks += (clk_data.qSecs * TICKS_PER_QSEC_64);
  totalTicks += clk_data.ticks;
  return totalTicks;
}

/**
 * Converts a uint64_t tick total (raw data) into a ClockData struct.
 */
ClockData convertTickTotal(uint64_t totalTicks)
{
  ClockData c_data;
  uint64_t modTicks = totalTicks % TICKS_PER_DAY_64; //mod it down to 40 bits
  uint64_t min64 = modTicks / MIN_PER_DAY_64; //whole minutes
  modTicks -= (min64 * MIN_PER_DAY_64); //mod ticks left over <= 1 minute
  uint64_t qSecs64 = modTicks / QSECS_PER_MIN_64; //whole quarter seconds
  modTicks -= (qSecs64 * QSECS_PER_MIN_64); //mod ticks left over <= 1/4 sec
  
  c_data.minutes = (uint16_t) min64;
  c_data.qSecs = (uint8_t) qSecs64;
  c_data.ticks = (uint16_t) modTicks;
  return c_data;
}

void printClockData(ClockData c_data)
{
  uint16_t hours = c_data.minutes / 60;
  uint16_t properMin = c_data.minutes - (hours * 60);
  uint8_t qSecs = c_data.qSecs;
  uint8_t secs = 	qSecs / 4;
  qSecs -= (4 * secs);
  uint32_t micros = 250000UL * qSecs;
  micros += (4 * (uint32_t) c_data.ticks);
  MY_DEBUG("Clock data information\n");
  MY_DEBUG(hours);
  MY_DEBUG(" hr ");
  MY_DEBUG(properMin);
  MY_DEBUG(" min ");
  MY_DEBUG(secs);
  MY_DEBUG(" sec ");
  MY_DEBUG(micros);
  MY_DEBUG(" microsec (");
  MY_DEBUG(c_data.ticks);
  MY_DEBUG(" ticks)\n");
}
  
/**
 * There are 4 main granularities for timing here:
 *
 * 1) Ticks, ranging from 0 to 62,500 (1/4 second)
 * 2) Quarters, ranging from 0 to 240 (1 minute)
 * 3) Minutes, ranging from 0 to 1440 (1 day)
 * 4) Days, which get tracked on a standard calendar
 */
/*class ClockTimeStamp
{
  protected:
	 
	 ClockData clk_data;
	 
  public:
	 
	 ClockTimeStamp() 
	 { 
		setClockTime(0ULL); 
	 }
	 
	 ClockTimeStamp(uint64_t totalTicks) 
	 { 
		setClockTime(totalTicks);
	 }
	 
	 ClockTimeStamp(uint16_t min, uint8_t qsec, uint16_t ticks)
    {
      setClockTime(convertClockData(min, qsec, ticks));
    }
    
    ClockTimeStamp(*/
	 

#endif
