#include <stdio.h>
#include <string.h>
#include "stl_timertest.h"

U8 getHighestPos(U32 num)
{
  U32 bm[5] = {0x2, 0xC, 0xF0, 0xFF00, 0xFFFF0000};
  U8 sh[5] = {1, 2, 4, 8, 16};
  U8 pos = 0;
  for (S8 i = 4; i >= 0; i--)
  {
	 if (num & bm[i])
	 {
		num >>= sh[i];
		pos |= sh[i];
	 }
  }
  return pos;
}

/**
* Given a month from 1-12 and a year (so leap year can be determined),
* get the days in the given month.
*
* @param a_month the month
* @param a_year the year
* @return the number of days in the given month
*/
U8 cal_date::getDaysInMonth(U8 a_month, U16 a_year)
{
  static U8 monthDays[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  monthDays[1] += (a_year & 0x3) ? 0 : 1;
  return monthDays[a_month - 1];
}

/**
* Constructs a new cal_date object with the given month, day, and year.
*
* @param a_month the current month
* @param a_day the current day
* @param a_year the current year
* @return a new cal_date object initialized to the given parameters
*/
cal_date::cal_date(U8 a_month, U8 a_day, U16 a_year)
{
  month = a_month;
  day = a_day;
  year = a_year;
  dayInYear = getDaysSinceJan1(a_month, a_day, a_year);
}

/**
* Constructs a cal_date object with the given number of days since January 1
* and the given year.
*
* @param daysSinceJan1 the number of complete days since January 1
* @param a_year the current year
* @return a new cal_date object initialized to the given parameters
*/
cal_date::cal_date(U16 daysSinceJan1, U16 a_year)
{
  year = a_year;
  // 		static U8 monthDays[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  // 		monthDays[1] += (a_year & 0x3) ? 0 : 1;
  U16 dayStart = 0;
  U8 i = 0;
  for (i = 0; i < 12; i++)
  {
	 dayStart += getDaysInMonth(i+1, a_year);
	 if (daysSinceJan1 < dayStart) break;
  }
  dayStart -= getDaysInMonth(i+1, a_year);
  day = daysSinceJan1 - (dayStart - 1);
  month = i + 1;
  dayInYear = daysSinceJan1;
  /*		printf("Date %d days after 1/1: %d/%d\n", daysSinceJan1, month, day);*/
}

/**
* Constructs a cal_date object based on the given number of microseconds passed.
*
* @param a_microseconds the number of microseconds elapsed since 1/1/2010
* @return a cal_date object with the date
*/
cal_date::cal_date(U64 a_microseconds)
{
  U16 days = (U16) (a_microseconds / USECS_PER_DAY);
  U16 a_year = days % 365 + 2010;
  days = days - a_year * 365;
  cal_date(days, a_year);
}

/**
* Get the number of complete days that have passed since the 1st of the year,
* meaning that January 1 will return 0 and December 31 will return 364 or 365
* for a leap year.
*
* @param a_month the month (1-12)
* @param a_day the day (1-MAX_MONTH)
* @param a_year the year (2010)
* @return the number of days since January 1
*/
U16 cal_date::getDaysSinceJan1(U8 a_month, U8 a_day, U16 a_year)
{
  U8 startIndex = a_month - 1;
  U8 i;
  U16 dayTotal = 0;
  for (i = startIndex; i > 0; i--)
  {
	 dayTotal += getDaysInMonth(i, a_year);
  }
  dayTotal += a_day - 1;
  return dayTotal;
}

/**
* Get the current calendar date by translating the string that is introduced
* by the compiler.
*
* @return the current date
*/
cal_date cal_date::getMacroDate()
{
  static char monthNames[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", 
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
  char date_str[] = __DATE__;
  U8 i = 0;
  for (i = 0; i < 12; i++)
  {
	 if (strstr(date_str, monthNames[i])) break;
  }
  U8 month = i+1;
  U8 day = (date_str[4] == ' ') ? 0 : 10 * (date_str[4] - 0x30);
  day += (date_str[5] - 0x30);
  U16 year = (date_str[7] - 0x30) * 1000 + (date_str[8] - 0x30) * 100 +
  (date_str[9] - 0x30) * 10 + (date_str[10] - 0x30);
  
  /*		printf("Date: %d/%d/%d\n", month, day, year);*/
  return cal_date(month, day, year);
}

/**
* Takes the given date and calculates the total number of microseconds that
* have passed since January 1 at midnight.
*
* @return the total number of microseconds that have passed since 1/1/2010 0:00
*/
U64 cal_date::toMicroseconds()
{
  U64 dayInYearUsec = (U64) getDayInYear();
  dayInYearUsec *= USECS_PER_DAY;
  return dayInYearUsec;
}

/**
* Increment the day at midnight --
*/
void cal_date::incrementDay()
{
  U8 curDay = getDay();
  U8 curMonth = getMonth();
  U16 curYear = getYear();
  bool newMonth = (getDaysInMonth(curMonth, curYear) == curDay);
  if (newMonth)
  {
	 curMonth++;
	 curDay = 1;
	 if (curMonth == 13)
	 {
		curYear++;
		curMonth = 1;
	 }
  }
  day = curDay;
  month = curMonth;
  year = curYear;
  dayInYear = getDaysSinceJan1(day, month, year);
}

/**
* Constructs a new clk_time object with the given hour, minute, second, and
* microsecond.
*
* @param a_hour the current hour
* @param a_minute the current minute
* @param a_second the current second
* @param a_microsecond the current microsecond
* @return a new clk_time object initialized to the given parameters
*/
clk_time::clk_time(U8 a_hour, U8 a_minute, U16 a_second, U32 a_microsecond)
{
  hour = a_hour;
  minute = a_minute;
  second = a_second;
  microsecond = a_microsecond;
}

/**
* Constructs a new clk_time object with the given number of microseconds.
*
* @param a_microseconds the desired number of microseconds (0-86400e6)
* @return a new clk_time object initialized to the given parameters
*/
clk_time::clk_time(U64 a_microseconds)
{
  a_microseconds = a_microseconds % USECS_PER_DAY;
  
  U32 secs = (U32) (a_microseconds / 1000000ULL);
  microsecond = (U32) (a_microseconds - secs * 1000000ULL);
  
  U32 mins = secs / 60UL;
  second = (U8) (secs - mins * 60UL);
  
  U32 hrs = mins / 60UL;
  minute = (U8) (mins - hrs * 60UL);
  
  hour = (U8) hrs;
}

/**
* Gets a new clk_time based on the __TIME__ macro.
*/
clk_time clk_time::getMacroTime()
{
  char time_str[] = __TIME__;
  U8 hours = (time_str[0] - 0x30) * 10 + (time_str[1] - 0x30);
  U8 minutes = (time_str[3] - 0x30) * 10 + (time_str[4] - 0x30);
  U8 seconds = (time_str[6] - 0x30) * 10 + (time_str[7] - 0x30);
  // 		printf("%2d:%02d:%02d\n", hours, minutes, seconds);
  return clk_time(hours, minutes, seconds);
}

/**
* Converts the clk_time object to raw microseconds.
*
* @return the total number of microseconds represented by this object
*/
U64 clk_time::toMicroseconds()
{
  U64 total_usec = microsecond;
  U64 multiplier = 1000000ULL;
  total_usec += (second * multiplier);
  multiplier *= 60;
  total_usec += (minute * multiplier);
  multiplier *= 60;
  total_usec += (hour * multiplier);
  return total_usec;
}

int main()
{
  cal_date md4 = cal_date::getMacroDate();
  U64 usecs = md4.toMicroseconds();
  printf("Day of year: %d\n", md4.getDayInYear());
  clk_time mt5 = clk_time::getMacroTime();
  U64 usecsTime = mt5.toMicroseconds();
  printf("Microseconds: %llu\n", usecsTime + usecs);
  printf("Date: %u/%u/%u\tTime: %u:%02u:%02u\n", md4.getMonth(), md4.getDay(), md4.getYear(), mt5.getHour(), mt5.getMinute(), mt5.getSecond());
  
  clk_time nextSunset = clk_time(18, 6);
  clk_time nextSunrise = clk_time(6, 19);
  cal_date dateToday = cal_date::getMacroDate();
  U64 microsecsSunset = nextSunset.toMicroseconds() + dateToday.toMicroseconds();
  
  U64 microsecsSunrise = dateToday.toMicroseconds() + USECS_PER_DAY;
  microsecsSunrise += nextSunrise.toMicroseconds();
  
  U64 deltaT = microsecsSunrise - microsecsSunset;
  clk_time nightTime = clk_time(deltaT);
  printf("Difference in time (sunset-sunrise): %d:%02d:%02d\n", nightTime.getHour(), nightTime.getMinute(), nightTime.getSecond()); 
  
  return 0;
}

