
#ifndef _STL_TIMERTEST_H_
#define _STL_TIMERTEST_H_

// #include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fstream>
#include <iostream>

#define U8 	  	uint8_t
#define S8		int8_t
#define U16		uint16_t
#define S16		int16_t
#define U32		uint32_t
#define S32		int32_t
#define U64		uint64_t
#define S64		int64_t
#define F32		float

#define F_CPU (16000000UL)
#define F_MHZ ((F_CPU)/1000000UL)

#define USECS_PER_DAY (86400000000ULL)
#define USECS_PER_HOUR (3600000000ULL)
#define USECS_PER_MIN (60000000ULL)
#define USECS_PER_SEC (1000000ULL)

class cal_date
{
  protected:
	 
	 U8 month;
	 U8 day;
	 U16 year;
	 U16 dayInYear;
	 
  public:
	 
	 /**
	 * Getter for the month field.
	 */
	 U8 getMonth() { return month; }
	 
	 /**
	 * Getter for the day field.
	 */
	 U8 getDay() { return day; }
	 
	 /**
	 * Getter for the year field.
	 */
	 U16 getYear() { return year; }
	 
	 /**
	 * Getter for the dayInYear field.
	 */
	 U16 getDayInYear() { return dayInYear; }
	 
	 /**
	 * Given a month from 1-12 and a year (so leap year can be determined),
	 * get the days in the given month.
	 *
	 * @param a_month the month
	 * @param a_year the year
	 * @return the number of days in the given month
	 */
	 static U8 getDaysInMonth(U8 a_month, U16 a_year);
	 
	 /**
	 * Constructs a new cal_date object with the given month, day, and year.
	 *
	 * @param a_month the current month
	 * @param a_day the current day
	 * @param a_year the current year
	 * @return a new cal_date object initialized to the given parameters
	 */
	 cal_date(U8 a_month = 1, U8 a_day = 1, U16 a_year = 2010);
	 
	 /**
	 * Constructs a cal_date object with the given number of days since January 1
	 * and the given year.
	 *
	 * @param daysSinceJan1 the number of complete days since January 1
	 * @param a_year the current year
	 * @return a new cal_date object initialized to the given parameters
	 */
	 cal_date(U16 daysSinceJan1, U16 a_year = 2010);
	 
	 /**
	  * Constructs a cal_date object with the given number of microseconds since January 1
	  * and the given time.
	  *
	  * @param a_microseconds the number of microseconds elapsed since 1/1/2010 0:00
	  * @return a new cal_date object initialized to the given number of microseconds
	  */
	 cal_date(U64 a_microseconds);
	 
	 /**
	 * Get the number of complete days that have passed since the 1st of the year,
	 * meaning that January 1 will return 0 and December 31 will return 364 or 365
	 * for a leap year.
	 *
	 * @param a_month the month (1-12)
	 * @param a_day the day (1-MAX_MONTH)
	 * @param a_year the year (2010)
	 * @return the number of days since January 1
	 */
	 static U16 getDaysSinceJan1(U8 a_month, U8 a_day, U16 a_year);
	 
	 /**
	 * Get the current calendar date by translating the string that is introduced
	 * by the compiler.
	 *
	 * @return the current date
	 */
	 static cal_date getMacroDate();
	 
	 /**
	 * Takes the given date and calculates the total number of microseconds that
	 * have passed since January 1 at midnight.
	 *
	 * @return the total number of microseconds that have passed since 1/1/2010 0:00
	 */
	 U64 toMicroseconds();
	 
	 /**
	 * Increment the day at midnight --
	 */
	 void incrementDay();
};

class clk_time
{
  protected:
	 
	 U8 hour;
	 U8 minute;
	 U8 second;
	 U32 microsecond;
	 
  public:
	 
	 /**
	 * Getter for the hour field.
	 */
	 U8 getHour() { return hour; }
	 
	 /**
	 * Getter for the minute field.
	 */
	 U8 getMinute() { return minute; }
	 
	 /**
	 * Getter for the second field.
	 */
	 U8 getSecond() { return second; }
	 
	 /**
	 * Getter for the microsecond field.
	 */
	 U32 getMicrosecond() { return microsecond; }
	 
	 /**
	 * Constructs a new clk_time object with the given hour, minute, second, and
	 * microsecond.
	 *
	 * @param a_hour the current hour
	 * @param a_minute the current minute
	 * @param a_second the current second
	 * @param a_microsecond the current microsecond
	 * @return a new clk_time object initialized to the given parameters
	 */
	 clk_time(U8 a_hour = 0, U8 a_minute = 0, U16 a_second = 0, U32 a_microsecond = 0);
	 
	 /**
	 * Constructs a new clk_time object with the given number of microseconds.
	 *
	 * @param a_microseconds the desired number of microseconds (0-86400e6)
	 * @return a new clk_time object initialized to the given parameters
	 */
	 clk_time(U64 a_microseconds);
	 
	 /**
	 * Converts the clk_time object to raw microseconds.
	 *
	 * @return the total number of microseconds represented by this object
	 */
	 U64 toMicroseconds();
	 
	 /**
	 * Gets a new clk_time based on the __TIME__ macro.
	 */
	 static clk_time getMacroTime();
};

class time_stamp2
{
  protected:
	 
	 cal_date c_date;
	 clk_time c_time;
  
  public:
	 
	 time_stamp2()
	 {
		c_date = cal_date::getMacroDate();
		c_time = clk_time::getMacroTime();
	 }
	 
	 time_stamp2(U64 rawData)
	 {
		c_date = cal_date(rawData);
		rawData = rawData % USECS_PER_DAY;
		c_time = clk_time(rawData);
	 }
	 
	 U64 toMicroseconds()
	 {
		return c_date.toMicroseconds() + c_time.toMicroseconds();
	 }
	 
	 time_stamp2 add(U64 otherData)
	 {
		return time_stamp2(this->toMicroseconds() + otherData);
	 }
	 
	 time_stamp2 add(time_stamp2 otherData)
	 {
		U64 ts2_us = this->toMicroseconds() + otherData.toMicroseconds();
		time_stamp2 ts2_ts = time_stamp2(ts2_us);
		return ts2_ts;
	 }
	 
	 time_stamp2 add(clk_time otherTime)
	 {
		U64 ts2_us = this->toMicroseconds() + otherTime.toMicroseconds();
		time_stamp2 ts2_ts = time_stamp2(ts2_us);
		return ts2_ts;
	 }
	 
	 time_stamp2 add(U8 hours, U8 minutes, U8 seconds, U32 microseconds)
	 {
		clk_time ts2_ts = clk_time(hours, minutes, seconds, microseconds);
		return add(ts2_ts);
	 }
	 
	 time_stamp2 sub(U64 otherData)
	 {
		return time_stamp2(this->toMicroseconds() - otherData);
	 }
	 
	 time_stamp2 sub(time_stamp2 otherData)
	 {
		return time_stamp2(this->toMicroseconds() - otherData.toMicroseconds());
	 }
	 
	 time_stamp2 sub(U8 hours, U8 minutes, U8 seconds, U32 microseconds)
	 {
		return sub(clk_time(hours, minutes, seconds, microseconds));
	 }
	 
	 time_stamp2 sub(clk_time otherTime)
	 {
		U64 ts2_us = this->toMicroseconds() - otherTime.toMicroseconds();
		time_stamp2 ts2_ts = time_stamp2(ts2_us);
		return ts2_ts;
	 }
	 
	 U16 getYear() { return c_date.getYear(); }
	 U8 getMonth() { return c_date.getMonth(); }
	 U8 getDay() { return c_date.getDay(); }
	 U16 getDayInYear() { return c_date.getDayInYear(); }
	 U8 getHour() { return c_time.getHour(); }
	 U8 getMinute() { return c_time.getMinute(); }
	 U8 getSecond() { return c_time.getSecond(); }
	 U32 getMicrosecond() { return c_time.getMicrosecond(); }
	 
};
	 
  
#endif
