
#ifndef _EQUATION_UTILS_H_
#define _EQUATION_UTILS_H_

#include <math.h>

// #include "rs232int.h"
#include <fstream>
#include <iostream>
#include "Locations.h"
#include "Calendar.h"

// typedef struct
// {
//   uint8_t hours;
//   uint8_t minutes;
// } HourMin;

namespace Equations
{
  float convRadians(float angleDegrees)
  {
    return angleDegrees * (M_PI / 180.0f);
  }
  
  float convDegrees(float angleRadians)
  {
    return angleRadians * (180.0f / M_PI);
  }
  
  float getT(float univTime)
  {
    float univDays = univTime / 24.0f;
    uint8_t inputDate = Calendar::GetInputDate();
    uint8_t correctedMonth = Calendar::GetCorrectedMonth();
    uint16_t correctedYear = Calendar::GetCorrectedYear();
	 std::cout << "univDays: " << univDays << "; inputDate: " << inputDate+0 <<
		"; corrMonth: " << correctedMonth+0 << "; corrYear: " << correctedYear+0 << std::endl;
    float valMonth = floor(30.6f * correctedMonth + 0.5f);
    float valYear = floor(365.25 * (correctedYear - 1976));
    float valT = univDays + inputDate + valMonth + valYear - 8707.5;
    return valT / 36525.0;
  }
  
  float getC(float gVal)
  {
    float gRads = convRadians(gVal);
    float valC1 = 1.915f * sin(gRads);
    float valC2 = 0.020f * sin(2.0f * gRads);
    return valC1 + valC2;
  }
  
  float getAlpha(float lVal)
  {
    float lRads = convRadians(lVal);
    float valAlpha1 = 2.466f * sin(2.0f * lRads);
    float valAlpha2 = 0.053f * sin(4.0f * lRads);
    return lVal - valAlpha1 + valAlpha2;
  }
  
  float getDeclAngle(float obliqAngle, float alpha)
  {
    float obliqRad = convRadians(obliqAngle);
    float alphaRad = convRadians(alpha);
    float declRad = atan(tan(obliqRad) * sin(alphaRad));
    return convDegrees(declRad);
  }
  
  float getAltAngle(float declAngle, float latitude, float solarAngle)
  {
    float declRad = convRadians(declAngle);
    float solarRad = convRadians(solarAngle + 180.0f);
    float latRad = convRadians(latitude);
    float valAlt1 = sin(latRad) * sin(declRad);
    float valAlt2 = cos(latRad) * cos(declRad) * cos(solarRad);
    return convDegrees(asin(valAlt1 - valAlt2));
  }
  
  float getPreAzimuth(float declAngle, float latitude, float solarHourAngle, float altAngle)
  {
    float declRad = convRadians(declAngle);
    float solarRad = convRadians(solarHourAngle + 180.0f);
    float latRad = convRadians(latitude);
    float altRad = convRadians(altAngle);
    
    float valAzi1 = cos(latRad) * tan(declRad);
    float valAzi2 = sin(latRad) * cos(solarRad);
    float valAzi3 = cos(declRad) * (valAzi1 + valAzi2);
    float aziRad = acos(valAzi3 / cos(altRad));
    float preAzimuth = convDegrees(aziRad);
    return preAzimuth;
  }
  
  float getDaylightMin(float declAngle, float latitude, float elevation)
  {
    float declRad = convRadians(declAngle);
    float latRad = convRadians(latitude);
    float elevInMeters = elevation * 12.0f * 2.54f / 100.0f;
    float elevRad = convRadians(-0.8333f - 0.0347f * sqrt(elevInMeters));
    float riseVal1 = sin(elevRad);
    float riseVal2 = sin(latRad) * sin(declRad);
    float riseVal3 = cos(latRad) * cos(declRad);
    float riseRad = acos((riseVal1 - riseVal2) / riseVal3);
    return convDegrees(riseRad) * 4.0f;
  }
  
  float norm360(float number)
  {
    return number - (float) floor(number / 360.0f) * 360.0f;
  }
  
  float norm180(float number)
  {
    float newNum = norm360(number);
    return (newNum > 180.0f) ? newNum - 360.0f : newNum;
  }
  
  float norm24(float number)
  {
    return number - floor(number / 24.0) * 24;
  }
  
  char degSign()
  {
    return (char) 0xB0;
  }
}

HourMin getHourMin(float totalMin)
{
  HourMin hr_min;
  hr_min.hours = (uint8_t) floor(totalMin / 60.0f);
  hr_min.minutes = (uint8_t) floor(totalMin - 60.0f * hr_min.hours);
  return hr_min;
}

void run()
{
  using namespace std;
  char degSign = Equations::degSign();
  
  float longitude = Locations::MechLab::Longitude;
  float latitude = Locations::MechLab::Latitude;
  float offset = Locations::MechLab::GMTOffset;
  float elevation = Locations::MechLab::Elevation;
  float magDecl = Locations::MechLab::MagneticDecl;
  
  float utSLO = Calendar::GetInputHour() - offset;
  cout << "UTSLO: " << utSLO << endl;
  float tVal = Equations::getT(utSLO);
  cout << "T value: " << tVal << endl;
  
  float gVal = Equations::norm360(357.528 + 35999.05 * tVal);
  cout << "G value: " << gVal << endl;
  
  float cVal = Equations::getC(gVal);
  cout << "C value: " << cVal << endl;
  
  float lVal = Equations::norm360(280.46 + (tVal * 36000.77) + cVal);
  cout << "L value: " << lVal << endl;
  
  float alpha = Equations::getAlpha(lVal);
  float ghaVal = Equations::norm360(utSLO * 15 - 180 - cVal + lVal - alpha);
  
  float obliqAngle = 23.4393 - 0.013 * tVal;
  
  float declAngle = Equations::getDeclAngle(obliqAngle, alpha);
  cout << "Declination: " << declAngle << degSign << endl;
  
  float deltaEOT = (lVal - cVal - alpha) / 15.0;
  float totalSecs = deltaEOT * 3600;
  int8_t minEOT = (int8_t) floor(totalSecs / 60.0);
  int8_t secEOT = (int8_t) round(totalSecs - 60 * minEOT);
  int8_t sec10 = (secEOT / 10);
  int8_t sec1 = (secEOT % 10);
  cout << "Equation of time (min): " << minEOT << ":" << sec10 << sec1 << endl;
  
  float deltaLSOT = (-longitude + (15 * offset)) / 15 - deltaEOT;
  
  float solarHourAngle = Equations::norm180(ghaVal + longitude);
  float solarTime = Equations::norm24(12 + solarHourAngle / 15.0);
  
  float solarMin = Calendar::GetInputMinute() - deltaLSOT * 60.0;
  int8_t solarDay = (solarMin < 0) ? -1 : (solarMin >= 1440) ? 1 : 0;
  solarMin -= 1440 * solarDay;
  int16_t solarH = (int16_t) floor(solarMin / 60);
  int16_t solarM = (int16_t) floor(solarMin - 60.0 * solarH);
  char dayCh = (solarDay == -1) ? '-' : (solarDay == 1) ? '+' : '\0';
  cout << "Solar time: " << solarH << ":" << (solarM / 10) << (solarM % 10);
  if (dayCh != '\0')
  {
    cout << dayCh;
  }
  cout << endl;
  
  float timeMin = Calendar::GetInputMinute();
  int8_t timeDay = (timeMin < 0) ? -1 : (timeMin >= 1440) ? 1 : 0;
  timeMin -= 1440 * timeDay;
  int16_t timeH = (int16_t) floor(timeMin / 60);
  int16_t timeM = (int16_t) floor(timeMin - 60.0f * timeH);
  char timeCh = (timeDay == -1) ? '-' : (timeDay == 1) ? '+' : '\0';
  cout << "Clock time: " << timeH << ":" << (timeM / 10) << (timeM % 10);
  if (timeCh != '\0')
  {
    cout << timeCh;
  }
  cout << endl;
  
  float hourAngle = (solarMin - 720) / 4 * -1;
  cout << "Hour angle: " << hourAngle << degSign << endl;
  
  float altAngle = Equations::getAltAngle(declAngle, latitude, solarHourAngle);
  cout << "Altitude angle: " << altAngle << degSign << endl;
  
  float aziAngle = Equations::getPreAzimuth(declAngle, latitude, solarHourAngle, altAngle) - 180.0f;
  aziAngle *= (hourAngle > 0 && hourAngle < 180) ? -1 : 1;
  cout << "Azimuth angle: " << aziAngle << degSign << endl;
  
  float trueNorth = Equations::norm360(180.0f - aziAngle);
  float compHdg = trueNorth + magDecl;
  cout << "*** Ideal compass headings are: ***" << endl;
  cout << "From TRUE NORTH: " << trueNorth << degSign << endl;
  cout << "From MAGNETIC NORTH: " << compHdg << degSign << endl;
  
  float daylightMin = Equations::getDaylightMin(declAngle, latitude, elevation);
  float sunriseMin = 720 - daylightMin + 60 * deltaLSOT;
  HourMin riseHrMin = getHourMin(sunriseMin);
  cout << "Sunrise: " << riseHrMin.hours << ":" << riseHrMin.minutes / 10 << riseHrMin.minutes % 10 << endl;
  float sunsetMin = 720 + daylightMin + 60 * deltaLSOT;
  HourMin setHrMin = getHourMin(sunsetMin);
  cout << "Sunset: " << setHrMin.hours << ":" << setHrMin.minutes / 10 << setHrMin.minutes % 10 << endl;
  cout << endl;
}

#endif
