/*
 *  Copyright (C) 2010 Alex Chernetz <andy80586-AT-gmail-DOT-com>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import static java.lang.Math.abs;
import static java.lang.Math.PI;

/**
 * A collection of functions for fast math, converted to Java to facilitate
 * software debugging before transitioning to C/C++.
 *
 * @author achernet
 */
public class FastMath {

   public static float toRadians(float degAngle)
   {
      return degAngle * (float) Math.PI / 180.0f;
   }

   public static float toDegrees(float radAngle)
   {
      return radAngle * 180.0f / (float) Math.PI;
   }

   public static float sine2(float angleRad)
   {
      float P = 0.225f;
      float oneOverPi = 1.0f / (float) PI;

      angleRad = angleRad * oneOverPi;
      int k = (int) Math.round(angleRad);
      angleRad = angleRad - k;

      float theSine = (4 - 4 * abs(angleRad)) * angleRad;
      theSine = P * (theSine * abs(theSine) - theSine) + theSine;

      //Java doesn't like this:
      //return (k & 0x1) ? -theSine : theSine;
      int kOdd = k & 0x1;
      return (kOdd > 0) ? -theSine : theSine;
   }

   private static float sine(float angleRad) { return sine2(angleRad); }

   /**
    * The fast sine function from this thread:
    * <a href=http://www.devmaster.net/forums/showthread.php?t=5784></a>
    * 
    * @param angleRad the angle in radians
    * @param extraPrecision true if we want extra precision, else false
    * @return the sine of the angle
    */
   private static float sine(float angleRad, boolean extraPrecision)
   {
      float pi = (float) PI;
      float b = 4.0f / pi;
      float c = -4.0f / (pi * pi);
      float theSine = b * angleRad + c * angleRad * abs(angleRad);
      if (extraPrecision)
      {
         float p = 0.225f;
         theSine = p * (theSine * abs(theSine) - theSine) + theSine;
      }
      return theSine;
   }

   private static float cosine(float angleRad, boolean extraPrecision)
   {
      return sine(angleRad + (float) PI * 0.5f, extraPrecision);
   }

   public static float cosine2(float angleRad)
   {
      return sine2(angleRad + (float) PI * 0.5f);
   }

   private static float cosine(float angleRad) { return cosine(angleRad, true); }

   private static float tangent(float angleRad, boolean extraPrecision)
   {
      return sine(angleRad, extraPrecision) / cosine(angleRad, extraPrecision);
   }

   private static float tangent(float angleRad)
   {
      return sine(angleRad) / cosine(angleRad);
   }

   public static float tangent2(float angleRad)
   {
      return sine2(angleRad) / cosine2(angleRad);
   }

   private static float getTheta1(float r, boolean extraPrecision)
   {
      float piOver4 = (float) PI / 4.0f;
      return (extraPrecision) ? 0.1963f*r*r*r - 0.9817f*r + piOver4 :
         piOver4 - r*piOver4;
   }

   private static float getTheta2(float r, boolean extraPrecision)
   {
      float piOver4 = (float) PI / 4.0f;
      float pi3Over4 = piOver4 * 3.0f;
      return (extraPrecision) ? 0.1963f*r*r*r - 0.9817f*r + pi3Over4 :
         pi3Over4 - r*piOver4;
   }

   public static float arctan(float y, boolean extraPrecision)
   {
      float absY = abs(y);
      if (absY < (Math.PI / 12.0f)) return y;
      float k1 = (1 - absY) / (1 + absY);
      float angle = getTheta1(k1, extraPrecision);
      return (y < 0) ? -angle : angle;
   }

   public static float arctan(float r) { return arctan(r, true); }

   /**
    * Algorithm for fast {@link atan2} function. Source:
    * <a href=http://www.dspguru.com/dsp/tricks/fixed-point-atan2-with-self-normalization></a>
    *
    * @param y the Y coordinate
    * @param x the X coordinate
    * @param extraPrecision true if extra precision is desired, else false
    * @return the angle in radians
    */
   public static float arctan2(float y, float x, boolean extraPrecision)
   {
      float absY = abs(y) + 1.0e-10f;
      float r, angle;
      if (x >= 0)
      {
         r = ((x - absY) / (x + absY));
         angle = getTheta1(r, extraPrecision);
      }
      else
      {
         r = ((x + absY) / (absY - x));
         angle = getTheta2(r, extraPrecision);
      }
      return (y < 0) ? -angle : angle;
   }

   public static float arctan2(float y, float x) { return arctan2(y, x, true); }

   public static float arcsin(float x)
   {
      return arctan(x * invSqrt(1.0f - x * x));
   }

   public static float arccos(float x)
   {
      return (float) Math.PI * 0.5f - arcsin(x);
   }

   /**
    * The classic Carmack inverse square root algorithm.
    */
   public static float invSqrt(float x)
   {
      float xOver2 = 0.5f * x;
      int iflt = Float.floatToRawIntBits(x);
      iflt = 0x5f3759df - (iflt >> 1);
      x = Float.intBitsToFloat(iflt);
      return x * (1.5f - xOver2 * x * x);
   }

   /**
    * The classic Carmack square root algorithm.
    */
   public static float sqrt(float x)
   {
      float xOver2 = 0.5f * x;
      float y = x;
      int iflt = Float.floatToRawIntBits(y);
      iflt = 0x5f3759df - (iflt >> 1);
      y = Float.intBitsToFloat(iflt);
      y = y * (1.5f - (xOver2 * y * y));
      y = y * (1.5f - (xOver2 * y * y));
      return x * y;
   }
}
