/** \file Bits.h
 *    Library designed to facilitate bit operations and make code more readable
 *		for a modest sacrifice in performance.
 *
 *  Revisions:
 *    \li  ABC 02-02-10  Added some comments for the doxygen generator.
 *
 *  License:
 *    This file released under the GNU General Public License, version 3. The 
 *    program is intended for educational and other non-profit use only at
 *    this time.
 */
#ifndef _BITS_H_
#define _BITS_H_

#include <stdlib.h>
#include <avr/io.h>

/**
 * The Bits library establishes higher-level functions for working with
 * individual bits of data for the Atmel AVR microcontroller family. It
 * seeks to resolve issues with magic numbers making things hard to
 * troubleshoot in lab and causing unnecessary delays.
 *
 * All the methods here are <code>static</code>, so no constructor is
 * necessary at this time. One will be added, however, if I decide to
 * incorporate operator overloading (like for using array notation,
 * etc.) into this library.
 *
 * TODO: Extend the library to operate on bits in groups, such as the
 * WGM13/12/11/10 bits.
 *
 */
class Bits
{
   private:
      
		/**
		 * A quick lookup table for bitmasks, which should be faster than having
		 * to do the shifting arithmetic every time.
		 */
      static const uint8_t mBitLookupTable[8];
      
   public:
      
		/**
		 * Converts the given bit position into a bitmask to use with the 5
		 * functions below. It ANDs the least significant 3 bits to avoid the
		 * possibility of seg faulting while indexing the bitmask lookup table.
		 *
		 * @param pos the bit position, from 0 (the rightmost bit) to 7 (the 
		 *  leftmost bit)
		 * @return the bitmask indexed with the least 3 significant bits of
		 *  <code>pos</code>
		 */
		static uint8_t bit(uint8_t pos)
		{ 
			return Bits::mBitLookupTable[pos & 0x7];
		}
      
		/**
		 * Returns whether or not the bit masked by <code>bitMask</code> was
		 * a logic 1 or a logic 0.
		 *
		 * @param regValue the register value. Note that the macro retrieves this
		 *  first, then saves a copy to this parameter since these are volatile.
		 *  (Yes, this should be callable with a macro -- if it chokes, it's a
		 *  bug!)
		 * @param bitMask a bitmask that isolates a single bit to operate on.
		 * @return true if the bit was a logic 1, otherwise false for a logic 0.
		 */
      static bool getBit(uint8_t regValue, uint8_t bitMask)
      {
         return (regValue & bitMask);
      }
      
		/**
		 * Sets the bit indicated by <code>bitMask</code> to a logic 1.
		 *
		 * @see <code>getBit</code> for details on register macros
		 * @param regValue the register value
		 * @param bitMask a bitmask that isolates one bit
		 */
      static void setBit(uint8_t regValue, uint8_t bitMask)
      {
         regValue |= bitMask;
      }
		
      /**
		 * Sets the bit indicated by <code>bitMask</code> to a logic 0.
		 *
		 * @see <code>getBit</code> for details on register macros
		 * @param regValue the register value
		 * @param bitMask a bitmask that isolates one bit
		 */
      static void clearBit(uint8_t regValue, uint8_t bitMask)
      {
         regValue &= ~(bitMask);
      }
		
      /**
		 * Flips the bit indicated, either from 0 to 1 or from 1 to 0.
		 *
		 * @see <code>getBit</code> for details on register macros
		 * @param regValue the register value
		 * @param bitMask a bitmask that isolates one bit
		 */
      static void flipBit(uint8_t regValue, uint8_t bitMask)
      {
         regValue ^= bitMask;
      }
		
      /**
		 * Writes a logic 1 or 0 depending on <code>value</code>.
		 *
		 * @see <code>getBit</code> for details on register macros
		 * @param regValue the register value
		 * @param bitMask a bitmask that isolates one bit
		 * @param value true to write a logic 1, otherwise false
		 */
      static void writeBit(uint8_t regValue, uint8_t bitMask, bool value)
      {
         (value) ? setBit(regValue, bitMask) : clearBit(regValue, bitMask);
      }
};

/**
 * The 8 possible bitmasks -- 1 for each bit. Looking up a value in the table is faster
 * than doing bit shifting arithmetic.
 */
const uint8_t Bits::mBitLookupTable[8] = {1, 2, 4, 8, 0x10, 0x20, 0x40, 0x80};

/**
 * Class to organize operations involving IO ports on the AVR microcontroller
 * family.
 */
class Ports
{
   public:
      
		/**
		 * Return the correct address for the DDRx port, where x is the letter
		 * of the port from A to G.
		 *
		 * @param portLetter the letter that comes after "DDR". Can be capital
		 *  or lower case.
		 * @return the DDRx address with this letter, or DDRG if <code>
		 *  portLetter</code> was invalid.
		 */
      static uint8_t DDR(char portLetter)
      {
         switch (portLetter)
         {
            case 'A': case 'a': return DDRA;
            case 'B': case 'b': return DDRB;
            case 'C': case 'c': return DDRC;
            case 'D': case 'd': return DDRD;
            case 'E': case 'e': return DDRE;
            case 'F': case 'f': return DDRF;
            default: break;
         }
         return DDRG;
      }
      /**
		 * Return the correct address for the PINx port, where x is the letter
		 * of the port from A to G.
		 *
		 * @param portLetter the letter that comes after "PIN". Can be capital
		 *  or lower case.
		 * @return the PINx address with this letter, or PING if <code>
		 *  portLetter</code> was invalid.
		 */
      static uint8_t PIN(char portLetter)
      {
         switch (portLetter)
         {
            case 'A': case 'a': return PINA;
            case 'B': case 'b': return PINB;
            case 'C': case 'c': return PINC;
            case 'D': case 'd': return PIND;
            case 'E': case 'e': return PINE;
            case 'F': case 'f': return PINF;
            default: break;
         }
         return PING;
      }
		
      /**
		 * Return the correct address for the PORTx port, where x is the letter
		 * of the port from A to G.
		 *
		 * @param portLetter the letter that comes after "PORT". Can be capital
		 *  or lower case.
		 * @return the PORTx address with this letter, or PORTG if <code>
		 *  portLetter</code> was invalid.
		 */
      static uint8_t PORT(char portLetter)
      {
         switch (portLetter)
         {
            case 'A': case 'a': return PORTA;
            case 'B': case 'b': return PORTB;
            case 'C': case 'c': return PORTC;
            case 'D': case 'd': return PORTD;
            case 'E': case 'e': return PORTE;
            case 'F': case 'f': return PORTF;
            default: break;
         }
         return PORTG;
      }

		/**
		 * Sets one bit on an I/O port for analog mode, without activating the pull-up
		 * resistor. This causes the voltage on the pin to be floating. This might be
		 * useful for working with the ADC or another analog-based system, but it's the
		 * rarest-used mode (when it's on, it's more likely to be a mistake than an
		 * intentional setting, as was the case in Lab 4).
		 *
		 * @param port a capital or lower-case letter from A to F to indicate the port
		 * @param oneBit the bit to configure on the port, from 0 to 7.
		 */
      static void setAnalogFloating(char port, uint8_t oneBit)
      {
         Bits::writeBit(Ports::DDR(port), Bits::bit(oneBit), 0);
         Bits::writeBit(Ports::PORT(port), Bits::bit(oneBit), 0);
		}
      
		/**
		 * Sets one bit on an I/O port for analog mode with the pull-up resistor enabled.
		 * An example application for this can be found when enabling the DIAGA/B pins on 
		 * a given channel for the motor.
		 *
		 * TODO: Not sure if this is actually analog or digital? It's definitely an input
		 * but it might be a digital one...
		 *
		 * @param port a capital or lower-case letter from A to F to indicate the port
		 * @param oneBit the bit to configure on the port, from 0 to 7.
		 */
      static void setAnalogPullUp(char port, uint8_t oneBit)
      {
         Bits::writeBit(Ports::DDR(port), Bits::bit(oneBit), 0);
         Bits::writeBit(Ports::PORT(port), Bits::bit(oneBit), 1);
      }
      
		/**
		 * Sets one bit on an I/O port as a digital output, and sets the output to a logic high.
		 * Note that the PORT pins are ACTIVE-LOW when in digital mode!
		 * 
		 * @param port a capital or lower-case letter from A to F to indicate the port
		 * @param oneBit the bit to configure on the port, from 0 to 7.
		 */
      static void setDigitalHigh(char port, uint8_t oneBit)
      {
         Bits::writeBit(Ports::DDR(port), Bits::bit(oneBit), 1);
         Bits::writeBit(Ports::PORT(port), Bits::bit(oneBit), 0);
      }
		
		/**
		 * Sets one bit on an I/O port as a digital output, and sets the output to a logic low.
		 * Note that the PORT pins are ACTIVE-LOW when in digital mode!
		 * 
		 * @param port a capital or lower-case letter from A to F to indicate the port
		 * @param oneBit the bit to configure on the port, from 0 to 7.
		 */
      static void setDigitalLow(char port, uint8_t oneBit)
      {
         Bits::writeBit(Ports::DDR(port), Bits::bit(oneBit), 1);
         Bits::writeBit(Ports::PORT(port), Bits::bit(oneBit), 1);
      }
};

/*class BitGroup
{
	private:
		
		uint8_t mRegValue;
		
		uint8_t lowBitMask;
		
		uint8_t highBitMask;
		
	protected:
		
		BitGroup(uint8_t regValue, uint8_t lowBitMask, uint8_t highBitMask)
		{
			
	public:*/
		
#endif //_BITS_H_