/**
 * \file VNH3SP30WithBitsHeader.cpp
 *		This file implements the functions declared in motor.h for
 *		performing various operations with the VN3SP30 motor control chips 
 *      on a dual Me405 board, which uses the ATmega128 microcontroller.
 *
 *	Revisions:
 *		\li 01-25-10 ABC Initial implementation.
 *		\li 01-26-10 Final version for lab 3
 *	
 *	License:
 *		This file released under the LGPL.
 */

#include <stdlib.h>
#include <avr/io.h>
#include "Bits.h"
#include "VNH3SP30WithBitsHeader.h"

/** 
 * Create a motor object.
 *
 * @param channel the channel to start the motor on
 */
VNH3SP30WithBitsHeader::VNH3SP30WithBitsHeader()    //Constructor that initiallizes the objects
{
	channel1 = 1;
	set_channel(channel1);
	ICR1 = 2000;
	mode1 = 14;
	set_mode(mode1);
	dutycycle_percent1 = 0;
	set_dutycycle(dutycycle_percent1);
	set_brake(false);
	prescaler1 = 64;
	isClockwise1 = true;
	set_direction(isClockwise1);
	set_prescaler(prescaler1);
	TCCR1A |= (1 << COM1A1);
	TCCR1A |= (1 << COM1B1);
}

/** 
 * Set the WGM mode (WGM10, WGM11, WGM12, WGM13).
 *
 * @param wgm the mode, from 0-15 
 */
void VNH3SP30WithBitsHeader::set_mode(uint8_t wgm)
{
	if (wgm > 15) 
	{
		return;
	}
	TCCR1A &= ~((1 << WGM10) | (1 << WGM11));
	TCCR1B &= ~((1 << WGM12) | (1 << WGM13));
	if (wgm & 0x1)
	{
		TCCR1A |= (1 << WGM10);
	}
	if (wgm & 0x2)
	{
		TCCR1A |= (1 << WGM11);
	}
	if (wgm & 0x4)
	{
		TCCR1B |= (1 << WGM12);
	}
	if (wgm & 0x8)
	{
		TCCR1B |= (1 << WGM13);
	}
}

/**
 * Get the current WGM mode for Timer 1. Should return 14 for Fast PWM.
 *
 * @return the WGM mode for the timer
 */
uint8_t VNH3SP30WithBitsHeader::get_mode()
{
	uint8_t wgm = 0;
	if (TCCR1A & (1 << WGM10))
	{
		wgm += 1;
	}
	if (TCCR1A & (1 << WGM11))
	{
		wgm += 2;
	}
	if (TCCR1B & (1 << WGM12))
	{
		wgm += 4;
	}
	if (TCCR1B & (1 << WGM13))
	{
		wgm += 8;
	}
	return wgm;
}
/**
 * Sets the channel on which the motor is running.
 *
 * @param channeltype the motor channel -- either 1 or 2
 */
void VNH3SP30WithBitsHeader::set_channel(uint8_t channeltype)
{
	channeltype = 1;
	if (channeltype == 1)
	{
		channel1 = 1;
		DDRC &= ~(1 << 2);				//Turn on DIAGA and B
		PORTC |= (1 << 2);
		DDRD &= ~(3 << 5);
		DDRC |= 0b00000011;
		DDRB |= (1 << 6);
		DDRB &= ~(1 << 5);
	}
	else if (channeltype == 2)
	{
		channel1 = 2;
		DDRD &= ~(1 << 7);				//Turn on DIABA and B
		PORTD |= (1 << 7);
		DDRC &= ~(3 << 0);
		DDRD |= 0b01100000;
		DDRB |= (1 << 5);
		DDRB &= ~(1 << 6);
	}
}

/**
 * Returns the current channel on which the motor is running.
 *
 * @return the current motor channel
 */
uint8_t VNH3SP30WithBitsHeader::get_channel()
{
	return channel1;
}

/**
 * Sets the direction of the motor -- either clockwise or counterclockwise.
 *
 * @param isClockwise1 True if the motor should spin clockwise, otherwise false
 */
void VNH3SP30WithBitsHeader::set_direction(bool isClockwise)
{
	isClockwise1 = isClockwise;
	if (isClockwise1 == true) 			//spins clockwise
	{     
		switch (channel1)
		{
			case 1:
				PORTC &= ~(1 << 1);
				PORTC |= (1 << 0);
				break;
			case 2:
				PORTD &= ~(1 << 6);
				PORTD |= (1 << 5); 
				break;
			default: break;
		}
	}
	else 		//spins counter clockwise
	{
		switch(channel1)
		{
			case 1:
				PORTC &= ~(1 << 0);
				PORTC |= (1 << 1);
				break;
			case 2:
				PORTD &= ~(1 << 5);
				PORTD |= (1 << 6);
				break;
			default: break;
		}
	}
}

/**
 * Get the direction for the motor.
 *
 * @return true for clockwise, otherwise false
 */
bool VNH3SP30WithBitsHeader::get_direction(void)
{
	return isClockwise1;
}

/**
 * Applies a brake to deliver no torque, temporarily shutting off the direction
 * until the brake turns off.
 *
 * @param brake true to turn on the brake, otherwise false
 */
void VNH3SP30WithBitsHeader::set_brake(bool brake)
{
	if (brake == true) //Brake both connected to ground
	{
		switch (channel1) 
		{
			case 1:
				PORTC &= ~(0b00000011);
				break;
			case 2:
				PORTD &= ~(0b01100000);
				break;
			default: break;
		}
	}
	else //Restore the previous direction
	{
		set_direction(isClockwise1);
	}
}

/**
 * Sets the prescaler to 0 (off), 1, 8, 64, 256, or 1024.
 *
 * @param prescale_value the value of the prescale clock divider
 */
void VNH3SP30WithBitsHeader::set_prescaler(uint16_t prescale_value1)
{
	uint8_t cs_pins = compute_prescaler(prescale_value1);
	
	//AND out CS10, CS11, and CS12
	TCCR1B &= ~((1 << CS10) | (1 << CS11) | (1 << CS12));

	//Set prescaler on CS10/CS11/CS12 by ORing cs_pins
	if (cs_pins & 0x1)
	{
		TCCR1B |= (1 << CS10);
	}
	if (cs_pins & 0x2)
	{
		TCCR1B |= (1 << CS11);
	}
	if (cs_pins & 0x4)
	{
		TCCR1B |= (1 << CS12);
	}
}

/**
 * Returns the current clock divisor set on the prescaler.
 *
 * @return the current prescale value
 */
uint16_t VNH3SP30WithBitsHeader::get_prescaler(void)
{
	return prescaler1;
}

/**
 * Sets the duty cycle as a percentage from 0 to 100.
 *
 * @param dutycycle The duty cycle, from 0 to 100
 */
void VNH3SP30WithBitsHeader::set_dutycycle(int8_t dutycycle)
{
	dutycycle_percent1 = dutycycle;
	if(dutycycle >= 0)
	{
		isClockwise1 = true;
	}
	else 
	{
		isClockwise1 = false;
		dutycycle *= -1;
	}
	uint16_t ocrValue = compute_dutycycle(dutycycle);
	uint8_t theChannel = get_channel();
	if (theChannel == 1)
	{
		set_direction(isClockwise1);
		OCR1B = ocrValue;
	}
	else
	{
		set_direction(isClockwise1);
		OCR1A = ocrValue;
	}
}

/**
 * This method returns the current duty cycle. 
 *
 * @return the current duty cycle
 */
int8_t VNH3SP30WithBitsHeader::get_dutycycle()
{
	return dutycycle_percent1;
}

/**
 * Computes the value to store in OCR1A or OCR1B given the desired duty cycle
 * as a percentage from 0 to 100.
 * 
 * @param percent the duty cycle percentage, from 0 to 100
 * @return the output comparison value to write to the timer
 */
uint16_t VNH3SP30WithBitsHeader::compute_dutycycle(uint8_t percent)
{
	return 20 * percent;
}

/**
 * Takes the prescale value and finds the corresponding index for CS10, CS11,
 * and CS12 that can be written to the TCCR1B register. If there is no match,
 * it returns the default value of 3 (i.e., prescale value of 64).
 * 
 * @param prescale_value the prescale value to set
 * @return the corresponding value to write to the CS pins
 */
uint8_t VNH3SP30WithBitsHeader::compute_prescaler(uint16_t prescale_value1)
{
	uint16_t values[] = {0, 1, 8, 64, 256, 1024, 0, 0};
	for (uint8_t i = 0; i < 7; i++)
	{
		if (values[i] == prescale_value1)
		{
			return i; //if prescale_value is valid, get its index
		}
	}
	return 3; //default prescale is 64
}
	
// end VNH3SP30WithBitsHeader.cpp




