/**
 * This program has the following features:
 *
 * User interface:
 * - Interactive time and date adjustment
 * - Interactive coordinate adjustment
 *
 * Advanced features:
 * - Automatic determination of the day of the week
 * - Automatic Daylight Savings Time calculation
 * - DST disabled automatically when inside rectangle approximating Arizona
 * - Support for Pacific, Mountain, Central, and Eastern time zones
 * - Coordinate input accurate to the nearest ten-thousandth
 * - ncurses in raw mode emulates serial port terminal for testing
 *
 * @author Alexander Chernetz <achernet-AT-calpoly-DOT-edu>
 * @author Patrick Bernard <ptbernar-AT-calpoly-DOT-edu>
 * @version 0.2 (February 22, 2010 23:45 PST)
 */
#include <curses.h>
#include <stdlib.h>
#include <stdint.h>

//coordinates for the lab
float mLat1 = 35.3039605;
float mLong1 = -120.6635398;

//test 2
float mLat2 = 32.386666; //32 deg. 23' 12"
float mLong2 = -111.044; //111 deg. 2' 38.4"

//temporary placeholders for latitude and longitude values
float mLat;
float mLong;

//The hour, minute, and GMT offset associated with the current time.
uint8_t mHour = 0;
uint8_t mMin = 0;
int8_t mGMTOffset = -8;

//Months and maximum days for each (inputting February 29 in a non-leap year
//results in automatic change to March 1)
char* mMonthNames[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", 
	"Aug", "Sep", "Oct", "Nov", "Dec"};
uint8_t mMonthDays[12] = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

//The current month, day, and year saved as globals for now.
uint8_t mMonth = 2;
uint8_t mDay = 23;
uint16_t mYear = 2010;

//The current weekday as determined by dayOfWeek.
uint8_t mWeekday = 7;
char* mWeekdayNames[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

/**
 * True if this is a leap year (by Gregorian definition).
 * 
 * @param year the year to check
 * @return true if it's a leap year, otherwise false
 */
bool isLeapYear(uint16_t year)
{
	return (year % 100 == 0) ? !(year % 400) : !(year & 0x3);
}

/**
 * Gets the month table value given the current month and whether or not this 
 * is a leap year.
 *
 * See <a href=http://en.wikipedia.org/wiki/Calculating_the_day_of_the_week>
 * this Wikipedia article</a> for more information.
 *
 * @param month the current month, from 1 to 12
 * @param lyr true if the current year is a leap year
 * @return the value as read from the month table
 */
uint8_t getMonthTableNo(uint8_t month, bool lyr)
{
	static uint8_t mTab[14] = {6,2,0,3,3,6,1,4,6,2,5,0,3,5};
	uint8_t index = (lyr && month <= 2) ? month - 1 : month + 1;
	return mTab[index];
}

/**
 * Returns a number from 0 to 6 indicating the day of the week, with Sunday
 * being set equal to 0 and increasing for each subsequent day as seen in
 * mWeekdayNames.
 *
 * @param month the current month
 * @param day the current day
 * @param year the current year
 * @return the day of the week, from 0 (Sunday) to 6 (Saturday)
 */
uint8_t dayOfWeek(uint8_t month, uint8_t day, uint16_t year)
{
	uint8_t cent = year / 100;
	uint8_t centNo = cent & 0x3;
	centNo = 2 * (3 - centNo);
	uint8_t year2 = (uint8_t) (year - 100 * cent);
	uint8_t year2q4 = year2 >> 2;
	bool lyr = isLeapYear(year);
	uint8_t monthNo = getMonthTableNo(month, lyr);
	uint8_t dayNo = (centNo + year2 + year2q4 + monthNo + day) % 7;
	return dayNo;
}

/**
 * Returns the first character in a 3-letter time zone string,
 * such as "PST" for Pacific Standard Time (GMT - 8).
 *
 * @param timeOffset the offset from Greenwich Mean Time (GMT)
 * @return the first letter of the time zone (E, C, M, or P)
 */
int8_t offsetToTimePrefix(int8_t timeOffset)
{
	switch (timeOffset)
	{
		case -5:
			return 'E';
		case -6:
			return 'C';
		case -7:
			return 'M';
		case -8:
			return 'P';
		default: break;
	}
	return 0x30 + (timeOffset < 0) ? -timeOffset : timeOffset;
}

/**
 * Clear the screen and show the main menu.
 */
void showMainMenu()
{
	refresh();
	clear();
	printw("Solar Tracking User Interface (ncurses version 0.2) Main Menu\n");
	printw("Choose an option:\n\n");
	printw("\t- <backspace>: clear the screen\n");
	printw("\t- t: set the time\n");
	printw("\t- d: set the date\n");
	printw("\t- c: set the coordinates\n");
	printw("\t- z: set the time zone\n");
	printw("\t- p: print current settings\n");
	printw("\t- s: compute solar calculations\n");
	printw("\t- h: help\n");
	printw("\t- q: quit screen\n");
}

/**
 * Converts any upper-case letters into lower-case letters.
 *
 * @param ch the character to test for upper-case conversion
 * @return the lower-case character if it can be converted
 */
int8_t toLower8(int8_t ch)
{
	return (ch >= 'A' && ch <= 'Z') ? ch + 0x20 : ch;
}

/**
 * Does the same thing as the function above, but does not
 * require casting from the wide ncurses character type and
 * losing information in the process.
 *
 * @param ch the character to test for upper-case conversion
 * @return the lower-case character if it can be converted
 */
int16_t toLower16(int16_t ch)
{
	return (ch >= 'A' && ch <= 'Z') ? ch + 0x20 : ch;
}

/**
 * Prints the time in standard 24-hour format. Pads the minutes
 * with a leading zero between 0 and 9 minutes after a given hour.
 *
 * @param hour the hour of the day (0-23)
 * @param min the minute of the given hour (0-59)
 */
void printTime(uint8_t hour, uint8_t min)
{
	printw("\nCurrent time: %d:%.2d\n", hour, min);
}

/**
 * Prints the date in standard format, like "Feb 22, 2010".
 *
 * @param month the current month
 * @param day the current day
 * @param year the current year
 */
void printDate(uint8_t month, uint8_t day, uint16_t year)
{
	const char* monthName = mMonthNames[month - 1];
	printw("\nToday's date: %s %d, %d\n", monthName, day, year);
}

/**
 * Returns true if the given latitude and longitude lie outside
 * a rough approximation of the area of the state of Arizona
 * (this is for testing with Arizona data)
 *
 * @param latitude the latitude of the location to test
 * @param longitude the longitude of the location to test
 * @return true if the location has a good chance of being in AZ, else false
 */
bool hasDST(float latitude, float longitude)
{
	return !(latitude > 31.277139 && latitude < 37.039200 &&
		longitude > -115.027288 && longitude < -108.914338);
}

/**
 * Given the date, time, and location, determine whether or not daylight
 * savings time is currently active. Daylight savings starts on the 2nd
 * Sunday in March and ends on the 1st Sunday in November, in both cases,
 * after 2:00 in the morning.
 * 
 * @param hour the current hour of day
 * @param month the current month of the year
 * @param day the current day of the month
 * @param year the current year (the weekdays change with each year)
 * @param latitude the latitude of the location being tested
 * @param longitude the longitude of the location being tested
 * @return true if Daylight Savings Time is currently active here, else false
 */
bool getDST(uint8_t hour, uint8_t month, uint8_t day, uint16_t year, 
				 float latitude, float longitude)
{
	if (!hasDST(latitude, longitude))
	{
		return false;
	}
	if (month != 3 && month != 11)
	{
		return (month > 3 && month < 11);
	}
	uint8_t day1 = dayOfWeek(month, 1, year);
	uint8_t sun1 = (day1 == 0) ? 1 : (1 + 7 - day1);
	uint8_t dayDST = sun1 + ((month == 3) ? 7 : 0);
	return (day != dayDST) ? (day > dayDST) : (hour >= 2);
}

/**
 * Prints the absolute date, in the following format:
 *
 * Currently: Mon, Feb 22, 2010 23:34 PST
 */
void printAbsoluteDate()
{
	uint8_t weekDay = dayOfWeek(mMonth, mDay, mYear);
	bool curDST = getDST(mHour, mMonth, mDay, mYear, mLat, mLong);
	int8_t timePrefix = offsetToTimePrefix(mGMTOffset);
	const char* dayName = mWeekdayNames[weekDay];
	const char* monthName = mMonthNames[mMonth - 1];
	printw("Currently: %s, %s %d, %d %d:%.2d %c%cT\n", dayName, monthName, mDay, mYear, mHour, mMin, timePrefix, (curDST) ? 'D' : 'S');
}	

/**
 * Prints the coordinates, in decimal degrees and with the traditional
 * North/South/East/West qualifiers (i.e., no negative numbers).
 */
void printCoords()
{
	int8_t ns = (mLat > 0) ? 'N' : 'S';
	int8_t ew = (mLong > 0) ? 'E' : 'W';
	printw("\nCoords: %2.3f %c, %2.3f %c\n", mLat, ns, mLong, ew);
}

/**
 * Converts a time zone letter into a GMT offset. Returns 42 (an obviously
 * invalid offset) if the letter does not correspond to a recognized
 * timezone.
 *
 * @param ch the character read in
 * @return the GMT offset
 */
int8_t checkForTimeZoneName(int16_t ch)
{
	switch (toLower16(ch))
	{
		case 'p':
			return -8;
		case 'm':
			return -7;
		case 'c':
			return -6;
		case 'e':
			return -5;
		default: break;
	}
	return 42;
}

/**
 * Converts an ASCII digit into its numerical value.
 *
 * @param ch the ASCII digit
 * @return the numerical value, or -1 if not a digit
 */
int8_t toDigit(int16_t ch)
{
	uint8_t digit = ch - 0x30;
	if (digit >= 0 && digit <= 9) return digit;
	return -1;
}

/**
 * Default handler for invalid user input.
 */
void invalidInput()
{
	clear();
	printw("Invalid input, try again\n");
}

/**
 * Counts the number of digits the given number has.
 * 
 * @param number the number to count digits on
 * @return the number of digits contained in the number.
 */
uint8_t getDigits(uint16_t number)
{
	uint8_t pl = 1;
	for (pl = 1; number >= 10; number /= 10, pl++);
	return pl;
}

/**
 * Get a whole number in the range specified. This will loop infinitely until a
 * valid number is typed in and the user hits ENTER.
 *
 * If ENTER is typed without any numbers having been keyed in first, it will be
 * considered equal to an input of 0, which will be valid if and only if
 * <code>low</code> is 0.
 * 
 * @param low the lowest allowed number
 * @param high the highest allowed number
 * @return a number within the range specified.
 */
uint16_t getWholeNum(uint16_t low, uint16_t high)
{
	uint32_t number = 0; //32 bits so we can handle 6-digit invalid numbers
	int16_t ch;
	//printw("Enter a whole number (%d-%d): ", low, high);
	while (true)
	{
		ch = getch();
		if (ch == '\r' || ch == '\n') //user hit enter
		{
			if (number >= low && number <= high) //final input is valid
			{
				break;
			}
			//else, it's invalid and will jump to end
		} 
		else //user should have entered a digit
		{
			uint8_t digit = toDigit(ch);
			if (digit >= 0 && digit <= 9) //it was a valid number
			{
				number = number * 10 + digit;
				printw("%c", ch);
				if (number <= high) //still in range
				{
					continue;
				}
			}
			//else, we got some other invalid character
		}
		//Number is invalid if we get here (automatic redo)
		number = 0;
		invalidInput();
		printw("Enter a whole number (%d-%d): ", low, high);
	}
	printw("\n");
	return (uint16_t) number;
}

/**
 * For latitude, the maximum angle allowed is 90. Since the algorithm cannot
 * calculate sun angles properly at the poles, it is limited to 89 degrees.
 *
 * For longitude, the maximum angle allowed is 180, where the International
 * Date Line is most commonly defined.
 *
 * @param maxAngle the maximum angle allowed by the coordinate system
 * @return the coordinate input by the user
 */
float getCoord(uint16_t maxAngle)
{
	uint16_t maxValue = maxAngle - 1;
	printw("\nInteger part: ");
	uint16_t intPart = getWholeNum(0, maxValue);
	float degs = (float) intPart;
	printw("\nFractional part in ten-thousandths: ");
	uint16_t fracPart = getWholeNum(0, 9999);
	float frac = (float) fracPart * 1.0e-4;
	return degs + frac;
}

/**
 * Selects Pacific, Mountain, Central, or Eastern time.
 *
 * @return the GMT offset for the given time zone
 */
int8_t enterTimeZone()
{
	int8_t timeEntered = 0;
	int16_t ch;
	
	while (true)
	{
		printw("Choose from the following time zones:\n");
		printw("\t- P: Pacific Time (GMT - 8) (default)\n");
		printw("\t- M: Mountain Time (GMT - 7)\n");
		printw("\t- C: Central Time (GMT - 6)\n");
		printw("\t- E: Eastern Time (GMT - 5)\n");
		printw("Enter your choice: ");
		ch = getch();
		if (ch == '\r' || ch == '\n')
		{
			printw("P\n");
			return -8;
		}
		timeEntered = checkForTimeZoneName(ch);
		if (timeEntered < -12 || timeEntered > 12)
		{
			invalidInput();
			continue;
		}
		break;
	}
	printw("%c\n", ch);
	return timeEntered;
}

/**
 * Function to enter an hour of day from 0 to 23.
 * 
 * @return the hour of day input by the user
 */
uint8_t enterHour()
{
	printw("Enter the hour of day: ");
	uint8_t hour = (uint8_t) getWholeNum(0, 23);
	return hour;
}

/**
 * Function to enter the minutes in a given hour from 0 to 59.
 * 
 * @return the minutes in the current hour, as input by the user.
 */
uint8_t enterMin()
{
	printw("Enter the minute: ");
	uint8_t min = (uint8_t) getWholeNum(0, 59);
	return min;
}

/**
 * Function to govern entering all the parameters required to
 * determine what time of day it is, including the time zone,
 * the hour, and the minute.
 */
void enterTime()
{
	printw("Setting time zone, hour, and minute...\n");
	mGMTOffset = enterTimeZone();
	mHour = enterHour();
	mMin = enterMin();
	//startTimer();
}

/**
 * Function to have the user enter the month as a number from 1 to 12.
 *
 * @return the month of the year as specified by user input
 */
uint8_t enterMonth()
{
	printw("Enter the month: ");
	uint8_t month = (uint8_t) getWholeNum(1, 12);
	return month;
}

/**
 * Function to have the user enter the day as a number from 1 to the maximum
 * number of days in the given month.
 * 
 * @param curMonth the current month, for determining the maximum number of
 * days
 * @return the day as entered by the user
 */
uint8_t enterDay(uint8_t curMonth)
{
	printw("Enter the day: ");
	uint8_t day = (uint8_t) getWholeNum(1, mMonthDays[curMonth - 1]);
	return day;
}

/**
 * Function to have the user enter a year between 1995 and 2099 for which to
 * calculate the incident angles of the Sun.
 * 
 * @return uint16_t the year as a 4-digit number.
 */
uint16_t enterYear()
{
	printw("Enter the year: ");
	uint16_t year = (uint16_t) getWholeNum(1995, 2099);
	return year;
}

/**
 * Function to have the user enter an entire date, including the
 * month, day, and year.
 */
void enterDate()
{
	printw("Setting the date...\n");
	uint8_t month = enterMonth();
	uint8_t day = enterDay(month);
	uint16_t year = enterYear();
	if (month == 2 && day == 29 && !isLeapYear(year)) //2/29 -> 3/1 if not LY
	{
		month = 3;
		day = 1;
	}
	mMonth = month;
	mDay = day;
	mYear = year;
	mWeekday = dayOfWeek(month, day, year);
}

/**
 * Function through which a user enters the latitude, between 0 and 89.9999
 * degrees; the North and South part is added later.
 *
 * @return the latitude in decimal degrees
 */
float enterLatitude()
{
	printw("Enter the latitude, starting with the integer part: ");
	float lat = getCoord(90);
	int16_t ch;
	while (true)
	{
		printw("\nEnter \'N/+\' for North or \'S/-\' for South (default: North): ");
		ch = toLower16(getch());
		if (ch == '\r' || ch == '\n' || ch == 'n' || ch == '+')
		{
			printw("N\n");
			break;
		}
		if (ch == 's' || ch == '-')
		{
			printw("S\n");
			lat *= -1;
			break;
		}
		invalidInput();
	}
	return lat;
}

/**
 * Function through which a user enters the longitude, between 0 and 179.9999
 * degrees; the East and West part is added later.
 *
 * @return the latitude in decimal degrees
 */
float enterLongitude()
{
	printw("Enter the longitude, starting with the integer part: ");
	float lon = getCoord(180);
	int16_t ch;
	while (true)
	{
		printw("\nEnter \'E/+\' for East or \'W/-\' for West (default: West): ");
		ch = toLower16(getch());
		if (ch == '\r' || ch == '\n' || ch == 'e' || ch == '+')
		{
			printw("E\n");
			break;
		}
		if (ch == 'w' || ch == '-')
		{
			printw("W\n");
			lon *= -1;
			break;
		}
	}
	return lon;
}	

/**
 * Function to govern the entering of coordinates through the user interface.
 */
void enterCoords()
{
	float latitude = enterLatitude();
	float longitude = enterLongitude();
	mLat = latitude;
	mLong = longitude;
}

/**
 * Print the most current settings for testing and debug purposes.
 */
void printCurrentSettings()
{
	printCoords();
	printAbsoluteDate();
}

/**
 * Main function where everything is set up here.
 */
int main()
{
	int16_t ch;
	initscr(); //initialize ncurses
	cbreak(); //similar to raw()
	keypad(stdscr, TRUE); //enable use of the keypad
	noecho(); //don't echo user input until specifically told to print it
	showMainMenu(); //show the main menu (help) right away.
	while (true)
	{
		printw("\nEnter an option: ");
		ch = toLower16(getch());
		switch (ch)
		{
			case '\r': case '\n': //this is true when the user hit ENTER
				printw("\nENTER was pressed\n");
				break;
			case KEY_F(1): //test: this works at detecting F1-F12 key presses
				printw("\nF1 was pressed\n");
				break;
			case 't': //enter the current time
				printw("%c\n", ch);
				enterTime();
				clear(); //clear to avoid too many messages on the console at once
				printTime(mHour, mMin);
				break;
			case 'h': //display the main menu again (help)
				printw("%c\n", ch);
				clear(); //clear to avoid too many messages on the console at once
				showMainMenu();
				break;
			case 'q': //quit the program
				printw("%c\n", ch);
				break;
			case 'd': //enter the current date
				printw("%c\n", ch);
				enterDate();
				clear(); //clear to avoid too many messages on the console at once
				printDate(mMonth, mDay, mYear);
				break;
			case 'p': //print the current settings in high detail
				clear();
				printCurrentSettings();
				break;
			case 'z': //enter a new time zone
				printw("%c\n", ch);
				mGMTOffset = enterTimeZone();
				clear(); //clear to avoid too many messages on the console at once
				printw("\nNew time zone offset: %d\n", mGMTOffset);
				break;
			case 'c': //enter coordinates
				printw("%c\n", ch);
				enterCoords();
				clear(); //clear to avoid too many messages on the console at once
				printCoords();
				break;
			case 263: //this is true when the user hit BACKSPACE
				clear(); //clear to avoid too many messages on the console at once
				break;
			default: //some other random key value was detected
				printw("\nEntered some other key: %d\n", ch);
				break;
		}
		if (ch == 'q' || ch == 'Q') break;
	}
	clear();
	printw("Now quitting. Press any key to continue...\n");
	getch();
	endwin();
	return 0;
}




