
#ifndef _ADC_4_128_H
#define _ADC_4_128_H

#include <avr/io.h>

#define BIT_POS(bit) (1 << (bit))
#define SET_0(reg,bit) ((reg) &= ~(BIT_POS(bit)))
#define SET_1(reg,bit) ((reg) |= (BIT_POS(bit)))
#define IS_1(reg,bit) (((reg) & BIT_POS(bit))?true:false)
#define IS_0(reg,bit) (!(IS_1((reg),(bit))))

#define QUEUE_SIZE (8)

typedef struct
{
  uint8_t channel;
  uint16_t value;
} ADCChannelResult;

queue<uint8_t, uint8_t, QUEUE_SIZE> ADCChannelQueue;

/**
 * Performs linear search of queue to determine if the given channel
 * is already in the queue.
 */
bool queueContains(uint8_t channel)
{
  uint8_t i = 0;
  uint8_t queueSize = ADCChannelQueue.num_items();
  bool foundDuplicate = false;
  
  for (i = 0; i < queueSize; i++)
  {
    uint8_t next = ADCChannelQueue.get();
    if (next == channel)
    {
      foundDuplicate = true;
    }
    ADCChannelQueue.put(next);
  }
  return foundDuplicate;
}

/**
 * Returns true if the JTAG must be disabled due to a channel being over 3.
 */
bool queueRequiresJTAGDisable()
{
  uint8_t i = 0;
  uint8_t queueSize = ADCChannelQueue.num_items();
  bool requiresDisable = false;
  
  for (i = 0; i < queueSize; i++)
  {
    uint8_t next = ADCChannelQueue.get();
    if (next > 3)
    {
      requiresDisable = true;
    }
    ADCChannelQueue.put(next);
  }
  return requiresDisable;
}

/**
 * Returns true if the queue is empty (no more channels needing to be read,
 * so the ADC should be getting shut down.
 */
bool queueIsEmpty()
{
  return ADCChannelQueue.is_empty();
}

/**
 * Tries to add the given channel, returns true or false depending
 * on how successful it was.
 */
bool tryAddChannel(uint8_t next) 
{
  bool tryResult = !(ADCChannelQueue.put(next));
  return tryResult;
}

/**
 * Function to enable/disable the ADC.
 */
void setADCEnabled(bool isEnabled)
{
  (isEnabled) ? SET_1(ADCSRA, ADEN) : SET_0(ADCSRA, ADEN);
}

/**
 * Function to turn free-running mode on/off.
 */
void setFreeRunning(bool isFreeRunning)
{
  (isFreeRunning) ? SET_1(ADCSRA, ADFR) : SET_0(ADCSRA, ADFR);
}

/**
 * Set up prescaler to divide by 128 for 16M / 128 = 125 kHz).
 */
void setUpPrescaler()
{
  SET_1(ADCSRA, ADPS2);
  SET_1(ADCSRA, ADPS1);
  SET_1(ADCSRA, ADPS0);
}

/**
 * Clear the interrupt flag.
 */
void clearInterruptFlag() 
{ 
  SET_0(ADCSRA, ADIF);
}

void setUpInterrupts()
{
  SET_1(ADCSRA, ADIE);
}

/**
 * Sets up voltage reference to the AVCC voltage.
 */
void setUpVoltageReference()
{
  SET_0(ADMUX, REFS1);
  SET_0(ADMUX, REFS0);
  MY_DEBUG("Voltage reference: ");
  if (IS_1(ADMUX, REFS0))
  {
	if (IS_0(ADMUX, REFS1))
	{
	  MY_DEBUG("AVCC\n");
	}
	else
	{
	  MY_DEBUG("2.56V\n");
	}
  }
  else
  {
	if (IS_0(ADMUX, REFS1))
	{
	  MY_DEBUG("AREF\n");
	}
	else
	{
	  MY_DEBUG("<error: this is a bug!>\n");
	}

  }
}

/**
 * Sets up the channel reading queue.
 */
void setUpQueue()
{
  ADCChannelQueue = queue<uint8_t, uint8_t, 8>();
}

/**
 * General setup routine for the ADC.
 */
void setUpADC()
{
  //PORTF = 0;
  //DDRF = 0;
  setADCEnabled(true); 		//turn ADC on
  setFreeRunning(false); 	//set single conversion mode
  setUpVoltageReference(); 	//set voltage to AVCC
  setUpPrescaler();		//set up prescaler
  clearInterruptFlag(); 	//clear any interrupt flags that might be there
  //setUpInterrupts();
  setUpQueue();			//set up 8 channel queue
  //SET_0(SFIOR, ACME);
  //SET_0(SFIOR, PUD);
}

/**
 * Enables or disables the JTAG interface to allow more ADC ports.
 * Requires that the same value be written twice to the JTD bit in
 * order to take effect.
 */
void setJTAGEnabled(bool isEnabled)
{
  uint8_t newMCUCSR = MCUCSR;
  if (isEnabled)
  {
    //SET_0(newMCUCSR, JTD);
    //MCUCSR = newMCUCSR;
    //MCUCSR = newMCUCSR;
	SET_0(MCUCSR, JTD);
	SET_0(MCUCSR, JTD);
    
    MY_DEBUG("JTAG status: Enabled\n");
  }
  else
  {
    SET_1(MCUCSR, JTD);
	 SET_1(MCUCSR, JTD);
    //MCUCSR = newMCUCSR;
    //MCUCSR = newMCUCSR;
    
    MY_DEBUG("JTAG status: Disabled\n");
  }  
}

/**
 * Changes the ADC channel on the multiplexer register.
 */
void changeChannel(uint8_t channel)
{
  uint8_t oldChannel = 0;
  oldChannel += (IS_1(ADMUX, MUX2)) ? 4 : 0;
  oldChannel += (IS_1(ADMUX, MUX1)) ? 2 : 0;
  oldChannel += (IS_1(ADMUX, MUX0)) ? 1 : 0;
  (channel & 0x4) ? SET_1(ADMUX, MUX2) : SET_0(ADMUX, MUX2);
  (channel & 0x2) ? SET_1(ADMUX, MUX1) : SET_0(ADMUX, MUX1);
  (channel & 0x1) ? SET_1(ADMUX, MUX0) : SET_0(ADMUX, MUX0);
  SET_0(ADMUX, MUX3); //bit 3 is always 0
  SET_0(ADMUX, MUX4); //bit 4 is always 0
  
  //SET_0(DDRF, channel);
  //SET_0(PORTF, channel);
  
  MY_DEBUG("Channel changed from ");
  MY_DEBUG(oldChannel);
  MY_DEBUG(" to ");
  MY_DEBUG(channel);
  MY_DEBUG("\n");
}

/**
 * Enqueue the channel if it hasn't been already.
 */
void enqueueChannel(uint8_t channel)
{
  if (queueContains(channel) || channel > 7)
  {
    MY_DEBUG("Already have channel ");
    MY_DEBUG(channel);
    MY_DEBUG("!\n");
    
    return; //it's already there!
  }
  
  //At this point, it will be able to add the channel unless
  //the queue is full.
  if (queueIsEmpty())
  {
    MY_DEBUG("Channel queue was empty, turning ADC back on\n");
    setADCEnabled(true);
  }
  bool result = tryAddChannel(channel);
  if (!result)
  {
    MY_DEBUG("Can't add ");
    MY_DEBUG(result);
    MY_DEBUG(" to queue (Queue was full!)\n");
    
    return;
  }
  //Disable JTAG if channel > 3
  /*if (channel > 3)
  {
    MY_DEBUG("Channel exceeds 3, disabling JTAG\n");
    setJTAGEnabled(false);
  }*/
  //Set the channel to be an analog input
  //SET_0(DDRF, channel);
  //SET_0(PORTF, channel);
  MY_DEBUG("Channel ");
  MY_DEBUG(channel);
  MY_DEBUG(" added successfully!\n"); 
}

/**
 * Dequeue and process the next channel.
 */
ADCChannelResult readNextChannel()
{
  ADCChannelResult returnValue;
  if (queueIsEmpty())
  {
    returnValue.channel = 0;
    returnValue.value = 0;
    return returnValue;
  }
  uint8_t channel = ADCChannelQueue.get();
  uint16_t result;
  
  MY_DEBUG("Now reading from channel ");
  MY_DEBUG(channel);
  MY_DEBUG("\n");
  
  changeChannel(channel);
  for (volatile int i = 0; i < 10000; i++);
  SET_1(ADCSRA, ADSC); //start the reading
  uint8_t bitmask = (1 << ADSC);
  while (ADCSRA & bitmask);
  result = ADCW;
  clearInterruptFlag();
  
  returnValue.value = result;
  returnValue.channel = channel;
  
  MY_DEBUG("Channel ");
  MY_DEBUG(channel);
  MY_DEBUG(" had a reading of ");
  MY_DEBUG(result);
  MY_DEBUG(".\n");
  
  /*if (!queueRequiresJTAGDisable())
  {
    MY_DEBUG("JTAG can be turned back on\n");
    setJTAGEnabled(true);
  }*/
  
  //If no more channels to read, shut this thing down
  /*if (queueIsEmpty())
  {
    setADCEnabled(false);
  }*/
  
  return returnValue;
}

#endif
